// Copyright (C) 2006-20008 Anders Logg.
// Licensed under the GNU LGPL Version 2.1.
//
// First added:  2006-05-16
// Last changed: 2008-11-13

#ifndef __MESH_EDITOR_H
#define __MESH_EDITOR_H

#include <vector>
#include <dolfin/common/types.h>
#include "CellType.h"

namespace dolfin
{
  
  class Mesh;
  class Point;
  class Vector;
  
  /// A simple mesh editor for creating simplicial meshes in 1D, 2D and 3D.

  class MeshEditor
  {
  public:
    
    /// Constructor
    MeshEditor();
    
    /// Destructor
    ~MeshEditor();

    /// Open mesh of given cell type, topological and geometrical dimension
    void open(Mesh& mesh, CellType::Type type, uint tdim, uint gdim);

    /// Open mesh of given cell type, topological and geometrical dimension
    void open(Mesh& mesh, std::string type, uint tdim, uint gdim);
    
    /// Specify number of vertices
    void initVertices(uint num_vertices);

    /// Specify number of cells
    void initCells(uint num_cells);
    
    /// Set the finite element signature for the type of parametric mapping used for the local
    /// elements in the mesh
    void setMeshCoordFEsignature(const std::string FE_signature)
    {mcv_finite_element_signature = FE_signature;}

    /// Set the dofmap signature for the type of parametric mapping used for the local
    /// elements in the mesh
    void setMeshCoordDofMapsignature(const std::string dofmap_signature)
    {mcv_dof_map_signature = dofmap_signature;}
    
    /// Set higher order mesh coordinates
    void setMeshCoordinates(Vector& mesh_coord);
    
    /// Set boolean indicator inside MeshGeometry
    void setAffineCellIndicator(uint c, const std::string affine_str);
    
    /// Add vertex v at given point p
    void addVertex(uint v, const Point& p);

    /// Add vertex v at given coordinate x
    void addVertex(uint v, double x);

    /// Add vertex v at given coordinate (x, y)
    void addVertex(uint v, double x, double y);

    /// Add vertex v at given coordinate (x, y, z)
    void addVertex(uint v, double x, double y, double z);

    /// Add cell with given vertices
    void addCell(uint c, const std::vector<uint>& v);

    /// Add cell (interval) with given vertices
    void addCell(uint c, uint v0, uint v1);

    /// Add cell (triangle) with given vertices
    void addCell(uint c, uint v0, uint v1, uint v2);
    
    /// Add cell (tetrahedron) with given vertices
    void addCell(uint c, uint v0, uint v1, uint v2, uint v3);

    /// Close mesh, finish editing, and order entities locally
    void close(bool order=true);

  private:

    // Add vertex, common part
    void addVertexCommon(uint v, uint dim);

    // Add cell, common part
    void addCellCommon(uint v, uint dim);

    // Clear all data
    void clear();

    // The mesh
    Mesh* mesh;

    // Topological dimension
    uint tdim;
    
    // Geometrical (Euclidean) dimension
    uint gdim;
    
    // Number of vertices
    uint num_vertices;

    // Number of cells
    uint num_cells;

    // Next available vertex
    uint next_vertex;

    // Next available cell
    uint next_cell;

    // Temporary storage for local cell data
    std::vector<uint> vertices;
    
    // stores the type of mesh mapping,
    // i.e. strings that specify the type of finite element function that represents the local mesh mapping
    std::string mcv_finite_element_signature;
    std::string mcv_dof_map_signature;

  };

}

#endif
